import importlib


def unserialize_object(data):
    """Given a metadata dictionary object, form a MetadataModel class

    Args:
        data (dict): Metadata to unserialize
    Returns:
        (BaseMetadataModel) Unserialized object
    """

    # Get the class to be loaded
    #   Assume the base class if '@class' not present
    class_name = data.get('@class', 'dlhub_sdk.models.BaseMetadataModel')

    # Make sure it is from the correct package
    if not class_name.startswith('dlhub_sdk.models.'):
        raise AttributeError('Metadata class must be from the `dlhub_sdk.models package')

    # Get the desired metadata class
    components = class_name.split(".")
    module_name = ".".join(components[:-1])
    mod = importlib.import_module(module_name)
    output = getattr(mod, components[-1])

    # Instantiate it using the user-provided data
    return output.from_dict(data)



"""Utilities for validating against DLHub schemas"""
from jsonschema import Draft4Validator, RefResolver
import requests

_schema_repo = "https://raw.githubusercontent.com/DLHub-Argonne/dlhub_schemas/master/schemas/"


def validate_against_dlhub_schema(document, schema_name):
    """Validate a metadata document against one of the DLHub schemas

    Note: Requires an internet connection

    Args:
        document (dict): Document instance to be validated
        schema_name (string): Name of schema (e.g., "dataset" for validating datasets).
            For full list, see: https://github.com/DLHub-Argonne/dlhub_schemas
    Raises:
        (jsonschema.SchemaError) If the schema fails to validate
    """

    # Make the schema validator
    schema = requests.get("{}/{}.json".format(_schema_repo, schema_name)).json()
    validator = Draft4Validator(schema, resolver=RefResolver(_schema_repo, schema))

    # Test the document
    validator.validate(document)


def codemeta_to_datacite(metadata):
    """Generate datacite from codemeta metadata

    Args:
        metadata (dict): Codemeta metadata
    Returns:
        (dict) datacite-compatabile metadata
    """

    # Taken from: https://github.com/caltechlibrary/ames/blob
    #             /920c028f7fcceaa1baf3a645a37de587dd2dbc41/matchers/caltechdata.py#L61

    datacite = {}
    if 'author' in metadata:
        creators = []
        for a in metadata['author']:
            cre = {}
            cre['creatorName'] = a['familyName'] + ', ' + a['givenName']
            cre['familyName'] = a['familyName']
            cre['givenName'] = a['givenName']
            if '@id' in a:
                idv = a['@id']
                split = idv.split('/')
                idn = split[-1]
                cre['nameIdentifiers'] = [{
                    'nameIdentifier': idn, 'nameIdentifierScheme': 'ORCID',
                    'schemeURI': 'http://orcid.org'}]
                # Should check for type and remove hard code URI
            if 'affiliation' in a:
                cre['affiliations'] = [a['affiliation']]
                # Should check if can support multiple affiliations
            creators.append(cre)
        datacite['creators'] = creators
    if 'name' in metadata:
        datacite['titles'] = [{'title': metadata['name']}]
    if 'license' in metadata:
        # Assuming uri to name conversion, not optimal
        uri = metadata['license']
        name = uri.split('/')[-1]
        datacite['rightsList'] = [{'rights': name, 'rightsURI': uri}]
    if 'keywords' in metadata:
        sub = []
        for k in metadata['keywords']:
            sub.append({"subject": k})
        datacite['subjects'] = sub
    if 'funder' in metadata:
        # Kind of brittle due to limitations in codemeta
        fund_list = []
        grant_info = ''
        if 'funding' in metadata:
            grant_info = metadata['funding'].split(',')
        if isinstance(metadata['funder'], list):
            count = 0
            for funder in metadata['funder']:
                entry = {'funderName': funder['name']}
                if '@id' in funder:
                    element = {}
                    element['funderIdentifier'] = funder['@id']
                    element['funderIdentifierType'] = 'Crossref Funder ID'
                    entry['funderIdentifier'] = element
                if grant_info != '':
                    split = grant_info[count].split(';')
                    entry['awardNumber'] = {'awardNumber': split[0]}
                    if len(split) > 1:
                        entry['awardTitle'] = split[1]
                count = count + 1
        else:
            funder = metadata['funder']
            entry = {'funderName': funder['name']}
            if '@id' in funder:
                element = {}
                element['funderIdentifier'] = funder['@id']
                element['funderIdentifierType'] = 'Crossref Funder ID'
                entry['funderIdentifier'] = element
            if grant_info != '':
                split = grant_info[0].split(';')
                entry['awardNumber'] = {'awardNumber': split[0]}
                if len(split) > 1:
                    entry['awardTitle'] = split[1]
            fund_list.append(entry)

        datacite['fundingReferences'] = fund_list
    return datacite



"""This module contains tools for describing objects being published to DLHub."""

from itertools import zip_longest
from datetime import datetime
from six import string_types
from zipfile import ZipFile
from glob import glob
import json
import sys
import os
import re

from dlhub_sdk.version import __version__
from dlhub_sdk.utils.schemas import codemeta_to_datacite

name_re = re.compile(r'^\S+$')


class BaseMetadataModel:
    """Base class for models describing objects published via DLHub

    Covers information that goes in the :code:`datacite` block of the metadata file and
    some of the DLHub block.

    --> Using a MetadataModel <--

    There are many kinds of MetadataModel classes that each describe a different kind of object.
    Each of these different types are created using the :code:`create_model` operation
    (e.g., :code:`KerasModel.create_model('model.h5')`), but have different arguments depending
    on the type of object. For example, TensorFlow models only require the directory created
    when saving the model for serving but scikit-learn models require the pickle file, how
    the pickle was created (e.g., with joblib), and how many input features it requires.

    Once created, you will need to fill in additional details about the object to make it reusable.
    The MetadataModel classes attempt to learn as much about an object as possible automatically,
    but there is some information that must be provided by a human. To start, you must define a
    title and name for the object and are encouraged to provide an abstract describing the model
    and list any associated papers/websites that describe the model. You will fill plenty of
    examples for how to describe the models in the DLHub_containers repostiory. Some types of
    objects require data specific to their type (e.g., Python servables need a list of required
    packages). We encourage you to find examples for your specific type of object in the containers
    repository for inspiration and to see the Python documentation for each Metadata Model.

    The MetadataModel object can be saved using the `to_dict` operation and read back into
    memory using the `from_dict` method. We recommend you save your dictionary to disk in the
    JSON or yaml format, which will allow for manual edits to be made before submitting or
    resubmitting a object description.
    """

    def __init__(self):
        """
        Initialize a dataset record
        """

        # Populate initial fields
        self._output = {"datacite": {
            "creators": [],
            "titles": [{'title': None}],
            "publisher": "DLHub",
            "publicationYear": str(datetime.now().year),
            "identifier": {'identifier': '10.YET/UNASSIGNED', 'identifierType': 'DOI'},
            "descriptions": [],
            "fundingReferences": [],
            "relatedIdentifiers": [],
            "alternateIdentifiers": [],
            "rightsList": []
        }, "dlhub": {
            "version": __version__,
            "domains": [],
            "visible_to": ["public"],
            'name': None,
            'files': {}
        }}

    def __getitem__(self, item):
        return self._output[item]

    @property
    def name(self):
        """
        Get the name of the servable

        Returns:
            (string) Name of the servable
        """
        return self['dlhub']['name']

    @classmethod
    def create_model(cls, **kwargs):
        """Instantiate the metadata model.

        Takes in arguments that allow metadata describing a dataset to be autogenerated. For
        example, these could include options describing how to read a dataset from a CSV file or
        which class method to invoke on a Python pickle object.
        """
        return cls()

    def read_codemeta_file(self, directory=None):
        """Read in metadata from a codemeta.json file

        Args:
            directory (string): Path to directory contain the codemeta.json file
                (default: current working directory)
        """

        # If no directory, use the current
        if directory is None:
            directory = os.getcwd()

        # Load in the codemeta
        with open(os.path.join(directory, 'codemeta.json')) as fp:
            codemeta = json.load(fp)

        # Convert it to datacite and store it
        datacite = codemeta_to_datacite(codemeta)
        self._output["datacite"].update(datacite)

        return self

    def set_authors(self, authors, affiliations=list()):
        """Add authors to a dataset

        Args:
            authors ([string]): List of authors for the dataset.
                In format: "<Family Name>, <Given Name>"
            affiliations ([[string]]): List of affiliations for each author.
        """
        for author, aff in zip_longest(authors, affiliations, fillvalue=[]):
            # Get the authors
            temp = author.split(",")
            family = temp[0].strip()
            given = temp[1].strip()

            # Add them to the list
            self._output["datacite"]["creators"].append({
                "givenName": given,
                "familyName": family,
                "affiliations": aff,
            })
        return self

    def set_title(self, title):
        """Add a title to the dataset"""
        self._output["datacite"]["titles"][0]["title"] = title
        return self

    def set_version(self, version):
        """Set the version of this resource

        Args:
            version (string): Version number
        """
        self._output["datacite"]["version"] = str(version)
        return self

    def set_abstract(self, abstract):
        """Define an abstract for this artifact. Use for a high-level summary

        Args:
            abstract (string): Description of this artifact
        """
        abs_block = {'description': abstract, 'descriptionType': 'Abstract'}

        # Delete Abstract block if it already exists
        self._output["datacite"]["descriptions"] = [
            x for x in self._output["datacite"]["descriptions"]
            if x["descriptionType"] != "Abstract"
        ]

        # Add the new block
        self._output["datacite"]["descriptions"].append(abs_block)
        return self

    def set_methods(self, methods):
        """Define a methods section for this artifact. Use to describe any specific details
        about how the dataset, model, etc was generated.

        Args:
            methods (str): Detailed method descriptions
        """
        method_blcok = {'description': methods, 'descriptionType': 'Methods'}

        # Delete Methods block if it already exists
        self._output["datacite"]["descriptions"] = [
            x for x in self._output["datacite"]["descriptions"]
            if x["descriptionType"] != "Methods"
        ]

        self._output["datacite"]["descriptions"].append(method_blcok)
        return self

    def set_domains(self, domains):
        """Set the field of science that is associated with this artifcat

        Args:
            domains ([string]): Name of a fields of science (e.g., "materials science")
        """
        self._output["dlhub"]["domains"] = domains
        return self

    def set_visibility(self, visible_to):
        """Set the list of people this artifact should be visible to.

        By default, it will be visible to anyone (["public"]).

        Args:
            visible_to ([string]): List of allowed users and groups, listed by GlobusAuth UUID
        """
        self._output["dlhub"]["visible_to"] = visible_to
        return self

    def set_doi(self, doi):
        """Set the DOI of this object, if available

        This function is only for advanced usage. Most users of the toolbox will not
        know the DOI before sending the doi in to DLHub.

        Args:
            doi (string): DOI of the object
        """
        self._output["datacite"]["identifier"]["identifier"] = doi
        return self

    def set_name(self, name):
        """Set the name of artifact.

        Should be something short, descriptive, and memorable

        Args:
            name (string): Name of artifact
        """
        if name_re.match(name) is None:
            raise ValueError('Name cannot contain any whitespace')
        self._output["dlhub"]["name"] = name
        return self

    def set_publication_year(self, year):
        """Define the publication year

        This function is only for advanced usage. Normally, this will be assigned automatically

        Args:
            year (string): Publication year
        """
        self._output["datacite"]["publicationYear"] = str(year)
        return self

    def add_rights(self, uri=None, rights=None):
        """Any rights information for this resource. Provide a rights management statement for the
        resource or reference a service providing such information. Include embargo information if
        applicable. Use the complete title of a license and include
        version information if applicable.

        Args:
            uri (string): URI of the rights
            rights (string): Description of the rights
        """
        if uri is None and rights is None:
            raise ValueError('You must defined either a URI or the rights')

        # Make the rights
        new_rights = {}
        if uri is not None:
            new_rights['rightsURI'] = uri
        if new_rights is not None:
            new_rights['rights'] = rights

        # Add it to the list
        self._output["datacite"]["rightsList"].append(new_rights)
        return self

    def add_funding_reference(self, name, identifier=None, identifier_type=None,
                              award_number=None, award_title=None, award_uri=None):
        """Add a funding source to the list of resources

        Args:
            name (string): Name of funding provider
            identifier (string): Identifier (e.g., ISNI) of the funder
            identifier_type (string): Type of the identifier (ISNI, GRID, Crossref Funder ID, Other)
            award_number (string): Code assigned by the funder
            award_title (string): Title of the award
            award_uri (string): URI of the award
        """

        # Error checking
        if identifier_type is None and identifier_type is not None:
            raise ValueError('Identifier type must be provided. "Other" is an option')
        if identifier_type not in ['ISNI', 'GRID', 'Crossref Funder ID', 'Other']:
            raise ValueError('Identifier type not recognized')

        # Format in datacite format
        funder = {'funderName': name}
        if identifier is not None:
            funder['funderIdentifier'] = {'funderIdentifier': identifier,
                                          'funderIdentifierType': identifier_type}
        if award_number is not None:
            funder['awardNumber'] = {'awardNumber': award_number}
            if award_uri is not None:
                funder['awardNumber']['awardURI'] = award_uri
        if award_title is not None:
            funder['awardTitle'] = award_title

        # Append to the total list
        self._output["datacite"]["fundingReferences"].append(funder)

        return self

    def add_alternate_identifier(self, identifier, identifier_type):
        """Add an identifier of this artifact in another service

        Args:
            identifier (string): Identifier
            identifier_type (string): Identifier type
        """

        self._output["datacite"]["alternateIdentifiers"].append({
            'alternateIdentifier': identifier,
            'alternateIdentifierType': identifier_type
        })
        return self

    def add_related_identifier(self, identifier, identifier_type, relation_type):
        """Add an identifier of an artifact that is related to this resource (e.g., a paper
        that describes a dataset).

        You must define both the identifier and how it relates to this resource. The possible types
        of relations are listed in the
        `documentation for datacite
        <https://schema.datacite.org/meta/kernel-4.1/doc/DataCite-MetadataKernel_v4.1.pdf>_`
        on Page 25. The most common one used in DLHub will likely be:

            - "IsDescribedBy": For a paper that describes a dataset or model
            - "IsDocumentedBy": For the software documentation for a model
            - "IsDerviedFrom": For the database a training set was pulled from

        Args:
            identifier (string): Identifier
            identifier_type (string): Identifier type
            relation_type (string): Relation between
        """

        if identifier_type not in ["ARK", "arXiv", "bibcode", "DOI", "EAN13", "EISSN", "Handle",
                                   "IGSN", "ISBN", "ISSN", "ISTC", "LISSN", "LSID", "PMID",
                                   "PURL", "UPC", "URL", "URN"]:
            raise ValueError('Unknown identifier type ({}).'.format(identifier_type))
        if relation_type not in ["IsCitedBy", "Cites", "IsSupplementTo", "IsSupplementedBy",
                                 "IsContinuedBy", "Continues", "IsNewVersionOf",
                                 "IsPreviousVersionOf", "IsPartOf", "HasPart", "IsReferencedBy",
                                 "References", "IsDocumentedBy", "Documents", "IsCompiledBy",
                                 "Compiles", "IsVariantFormOf", "IsOriginalFormOf",
                                 "IsIdenticalTo", "HasMetadata", "IsMetadataFor", "Reviews",
                                 "IsReviewedBy", "IsDerivedFrom", "IsSourceOf", "IsDescribedBy",
                                 "Describes"]:
            raise ValueError('Unknown relation type: ({})'.format(relation_type))

        self._output["datacite"]["relatedIdentifiers"].append({
            'relatedIdentifier': identifier,
            'relatedIdentifierType': identifier_type,
            'relationType': relation_type
        })
        return self

    def add_file(self, file, name=None):
        """Add a file to the list of files to be distributed with the artifact

        Args:
            file (string): Path to the file
            name (string): Optional. Name of the file, if it is a file that serves a specific
                purpose in software based on this artifact (e.g., if this is a pickle file
                of a scikit-learn model)
        """

        if name is None or name == "other":
            if "other" not in self._output["dlhub"]["files"]:
                self._output["dlhub"]["files"]["other"] = []
            self._output["dlhub"]["files"]['other'].append(file)
        else:
            self._output["dlhub"]["files"][name] = file
        return self

    def add_directory(self, directory, recursive=False):
        """Add all the files in a directory

        Args:
            directory (string): Path to a directory
            recursive (bool): Whether to add all files in a directory
        """

        # Get potential files
        if recursive:
            if sys.version_info < (3, 5):
                raise RuntimeError('You must use Python >= 3.5 to do recursive listing')
            hits = glob('{}/**/*'.format(directory), recursive=True)
        else:
            hits = glob('{}/*'.format(directory))

        # Get only the files
        files = [x for x in hits if os.path.isfile(x)]
        return self.add_files(files)

    def add_files(self, files):
        """Add files that should be distributed with this artifact.

        Args:
            files ([string]): Paths of files that should be published
        """

        # Type checking
        if isinstance(files, str):
            files = [files]

        # Add the files
        for file in files:
            self.add_file(file)
        return self

    def to_dict(self, simplify_paths=False, save_class_data=False):
        """Render the dataset to a JSON description

        Args:
            simplify_paths (bool): Whether to simplify the paths of each file
            save_class_data (bool): Whether to save data about the class
        Returns:
            (dict) A description of the dataset in a form suitable for download
        """

        # Check for required fields
        if self._output["datacite"]["titles"][0]["title"] is None:
            raise ValueError('Title must be specified. Use `set_title`')
        if self._output["dlhub"]["name"] is None:
            raise ValueError('Name must be specified. Use `set_name`')

        # Make a copy of the output
        out = dict(self._output)

        # Add the name of the class to the output, if desired
        if save_class_data:
            out['@class'] = '{}.{}'.format(self.__class__.__module__,
                                           self.__class__.__name__)

        # Prepare the files
        if simplify_paths:
            # Get the common path
            common_path = self._get_common_path()

            files = {}
            for k, v in out["dlhub"]["files"].items():
                if k == "other":
                    files[k] = [os.path.relpath(f, common_path) for f in v]
                else:
                    files[k] = os.path.relpath(v, common_path)

            # Copy over the current files list
            out["dlhub"]["files"] = files

        return out

    @classmethod
    def from_dict(cls, data):
        """Reconstitute class from dictionary

        Args:
            data (dict): Metadata for this class
        """

        # Create the object, overwrite data
        output = cls()
        output._output = dict(data)

        # Make sure the class information is removed
        if '@class' in output._output:
            del output._output['@class']

        return output

    def list_files(self):
        """Provide a list of files associated with this artifact.

        Returns:
            ([string]) list of file paths"""
        # Gather a list of all the files
        output = []
        for k, v in self._output["dlhub"]["files"].items():
            if isinstance(v, string_types):
                output.append(v)
            else:  # It is a list
                output.extend(v)
        return output

    def get_zip_file(self, path):
        """Write all the listed files to a ZIP object

        Takes all of the files returned by `list_files`. First determines the largest common
        path of all files, and preserves directory structure by using this common path as the
        root directory. For example, if the files are "/home/a.pkl" and "/home/a/b.dat", the common
        directory is "/home" and the files will be stored in the Zip as "a.pkl" and "a/b.dat"

        Args:
            path (string): Path for the ZIP File
        Returns:
            (string): Base path for the ZIP file (useful for adjusting the paths of the files
                included in the metadata model)
        """

        # Open the zip file in "exclusively create" (x) mode
        with ZipFile(path, 'x') as newzip:
            if len(self.list_files()) == 0:
                return "."

            # Get the common path of all files
            root_path = self._get_common_path()

            # Add each file to the directory
            for file in self.list_files():
                newzip.write(file, arcname=os.path.relpath(file, root_path))

            return root_path

    def _get_common_path(self):
        """Determine the common path of all files

        Returns:
            (string) Common path
        """
        # Get the files
        files = self.list_files()

        # Shortcut: if no files
        if len(files) == 0:
            return '.'

        # Get the directories for all the files
        directories = [f if os.path.isdir(f) else os.path.dirname(f) for f in files]

        # Get the largest common path
        return os.path.commonpath(os.path.abspath(d) for d in directories)



import pandas as pd

from dlhub_sdk.models import BaseMetadataModel
from dlhub_sdk.utils.types import simplify_numpy_dtype


class Dataset(BaseMetadataModel):
    """Base class for describing a dataset

    The Dataset class and any of its subclasses contain operations
    for describing what a dataset is and how to use it.
    """

    def __init__(self):
        super(Dataset, self).__init__()

        # Add the datacite type as Dataset
        self._output['datacite']['resourceType'] = {'resourceTypeGeneral': 'Dataset'}

        # Initialize the "dataset" block
        self._output['dataset'] = {}


class TabularDataset(Dataset):
    """Read a dataset stored as a single file in a tabular format.

    Will read in the names of the columns, and allow users to associate
    column names with descriptions of the data provided.

    This class is compatible with any data format readable by the Pandas
    library. See the list of
    `read functions in Pandas <https://pandas.pydata.org/pandas-docs/stable/io.html>`_
    """

    @classmethod
    def create_model(cls, path, format="csv", read_kwargs=None):
        """Initialize the description of a tabular dataset

        Args:
            path (string): Path to dataset
            format (string): Format of the dataset. We support all of the
                read operations of Pandas (e.g., read_csv). Provide the format
                of your dataset as the suffix for the Pandas read command (e.g.,
                "csv" for "read_csv").
            read_kwargs (dict): Any keyword arguments for the pandas read command
        """
        output = cls()
        if read_kwargs is None:
            read_kwargs = {}
        output.load_dataset(path, format, **read_kwargs)
        return output

    def load_dataset(self, path, format, **kwargs):
        """Load in a dataset to get some high-level descriptions of it

        Args:
            path (string): Path to dataset
            format (string): Format of the dataset. We support all of the
                read operations of Pandas (e.g., read_csv). Provide the format
                of your dataset as the suffix for the Pandas read command (e.g.,
                "csv" for "read_csv").
            **kwargs (dict): arguments for the Pandas read function
        """

        # Add the data as the path of interest
        self.add_file(path, 'data')

        # Store the format information
        self._output["dataset"]["format"] = format
        self._output["dataset"]["read_options"] = kwargs

        # Read in the data
        read_fun = getattr(pd, 'read_{}'.format(format))
        data = read_fun(path, **kwargs)
        self._output["dataset"]["columns"] = [
            {'name': c, 'type': simplify_numpy_dtype(d)}
            for c, d in zip(data.columns, data.dtypes)
        ]

        # Zero out the input and output columns
        for x in ["inputs", "labels"]:
            if x in self._output["dataset"]:
                del self._output["dataset"][x]

    def annotate_column(self, column_name, description=None, data_type=None, units=None):
        """Provide documentation about a certain column within a dataset.

        Overwrites any type values inferred from reading the dataset

        Args:
            column_name (string): Name of a column
            description (string): Longer description of a column
            data_type (string): Short description of the data type
            units (string): Units for the columns data (if applicable)
        """
        column = self._get_column(column_name)
        if description is not None:
            column['description'] = description
        if data_type is not None:
            column['type'] = data_type
        if units is not None:
            column['units'] = units
        return self

    def get_unannotated_columns(self):
        """Get the names of columns that have not been described"""

        return [x["name"] for x in self["dataset"]["columns"] if "description" not in x]

    def _get_column(self, column_name):
        """Gets the metadata for a certain column

        Args:
            column_name (string): Name of column to be altered
        Returns:
            (dict) Column metadata. The editable object, not a copy
            """
        for column in self._output["dataset"]["columns"]:
            if column["name"] == column_name:
                return column
        raise ValueError('No such column {}'.format(column_name))

    def mark_inputs(self, column_names):
        """Mark which columns are inputs to a model

        Args:
            column_names ([string]): Names of columns
        """
        # Make sure all the columns exist
        for c in column_names:
            self._get_column(c)
        self._output["dataset"]["inputs"] = list(column_names)
        return self

    def mark_labels(self, column_names):
        """Mark a column as label

        Args:
            column_names ([string]): Names of columns
        """
        for c in column_names:
            self._get_column(c)
        self._output["dataset"]["labels"] = list(column_names)
        return self



import json
import os
from tempfile import mkstemp

from globus_sdk.base import BaseClient, slash_join
import jsonpickle
import mdf_forge
import mdf_toolbox
import pandas as pd
import requests

from dlhub_sdk.config import DLHUB_SERVICE_ADDRESS, CLIENT_ID, SEARCH_INDEX
from dlhub_sdk.utils.schemas import validate_against_dlhub_schema


class DLHubClient(BaseClient):
    """Main class for interacting with the DLHub service

    Holds helper operations for performing common tasks with the DLHub service. For example,
    `get_servables` produces a list of all servables registered with DLHub.

    For most cases, we recommend creating a new DLHubClient by calling ``DLHubClient.login``.
    This operation will check if you have saved any credentials to disk before using the CLI or SDK
    and, if not, get new credentials and save them for later use.
    For cases where disk access is unacceptable, you can create the client by creating an authorizer
    following the
    `tutorial for the Globus SDK <https://globus-sdk-python.readthedocs.io/en/stable/tutorial/>_
    and providing that authorizer to the initializer (e.g., ``DLHubClient(auth)``)"""

    def __init__(self, dlh_authorizer=None, search_client=None, http_timeout=None,
                 force_login=False, **kwargs):
        """Initialize the client

        Args:
            dlh_authorizer (:class:`GlobusAuthorizer
                            <globus_sdk.authorizers.base.GlobusAuthorizer>`):
                An authorizer instance used to communicate with DLHub.
                If ``None``, will be created.
            search_client (:class:`SearchClient <globus_sdk.SearchClient>`):
                An authenticated SearchClient to communicate with Globus Search.
                If ``None``, will be created.
            http_timeout (int): Timeout for any call to service in seconds. (default is no timeout)
            force_login (bool): Whether to force a login to get new credentials.
                A login will always occur if ``dlh_authorizer`` or ``search_client``
                are not provided.
        Keyword arguments are the same as for BaseClient.
        """
        if force_login or not dlh_authorizer or not search_client:
            auth_res = mdf_toolbox.login(services=["search", "dlhub"], app_name="DLHub_Client",
                                         client_id=CLIENT_ID, clear_old_tokens=force_login,
                                         token_dir=os.path.expanduser("~/.dlhub/credentials"))
            dlh_authorizer = auth_res["dlhub"]
            search_client = auth_res["search"]
        # Unused variable, will be used in future
        self.__forge_client = mdf_forge.Forge(index=SEARCH_INDEX, services=[],  # noqa: F841
                                              clients={"search": search_client})
        super(DLHubClient, self).__init__("DLHub", environment='dlhub', authorizer=dlh_authorizer,
                                          http_timeout=http_timeout, base_url=DLHUB_SERVICE_ADDRESS,
                                          **kwargs)

    @classmethod
    def login(cls, force=False, **kwargs):
        """Create a DLHubClient with credentials

        Either uses the credentials already saved on the system or, if no credentials are present
        or ``force=True``, runs a login procedure to get new credentials

        Keyword arguments are passed to the DLHubClient constructor

        Args:
            force (bool): Whether to force a login to get new credentials.
        Returns:
            (DLHubClient) A client complete with proper credentials.
        """
        auth_res = mdf_toolbox.login(services=["search", "dlhub"], app_name="DLHub_Client",
                                     token_dir=os.path.expanduser("~/.dlhub/credentials"),
                                     clear_old_tokens=force)
        return DLHubClient(dlh_authorizer=auth_res["dlhub"], search_client=auth_res["search"],
                           **kwargs)

    def _get_servables(self):
        """Get all of the servables available in the service

        Returns:
            (pd.DataFrame) Summary of all the models available in the service
        """

        r = self.get("servables")
        return pd.DataFrame(r.data)

    def get_servables(self):
        """Get all of the servables available in the service

        This is for backwards compatibility. Previous demos relied on this function
        prior to it being made an internal function.

        Returns:
            (pd.DataFrame) Summary of all the models available in the service
        """

        return self._get_servables()

    def list_servables(self):
        """Get a list of the servables available in the service

        Returns:
            (pd.DataFrame) Summary of all the models available in the service
        """
        df_tmp = self._get_servables()
        return df_tmp[['dlhub_name']]

    def get_task_status(self, task_id):
        """Get the status of a DLHub task.

        Args:
            task_id (string): UUID of the task
        Returns:
            (dict) status block containing "status" key.
        """

        r = self.get("{task_id}/status".format(task_id=task_id))
        return r.json()

    def describe_servable(self, author, name):
        """
        Get the description for a certain servable

        Args:
            author (string): Username of the owner of the servable
            name (string): Name of the servable
        Returns:
            (pd.DataFrame) Summary of the servable
        """

        df_tmp = self._get_servables()

        # Downselect to more useful information
        df_tmp = df_tmp[['name', 'description', 'input', 'output', 'author', 'status']]

        # Get the desired servable
        serv = df_tmp.query('name={name} AND author={author}'.format(name=name, author=author))
        return serv.iloc[0]

    def run(self, name, inputs, input_type='python'):
        """Invoke a DLHub servable

        Args:
            name (string): DLHub name of the model of the form <user>/<model_name>
            inputs: Data to be used as input to the function. Can be a string of file paths or URLs
            input_type (string): How to send the data to DLHub. Can be "python" (which pickles
                the data), "json" (which uses JSON to serialize the data), or "files" (which
                sends the data as files).
        Returns:
            Reply from the service
        """
        servable_path = 'servables/{name}/run'.format(name=name)

        # Prepare the data to be sent to DLHub
        if input_type == 'python':
            # data = {'python': codecs.encode(pkl.dumps(inputs), 'base64').decode()}
            data = {'python': jsonpickle.encode(inputs)}
        elif input_type == 'json':
            data = {'data': inputs}
        elif input_type == 'files':
            raise NotImplementedError('Files support is not yet implemented')
        else:
            raise ValueError('Input type not recognized: {}'.format(input_type))

        # Send the data to DLHub
        r = self.post(servable_path, json_body=data)
        if r.http_status != 200:
            raise Exception(r)

        # Return the result
        return r.data

    def publish_servable(self, model):
        """Submit a servable to DLHub

        If this servable has not been published before, it will be assigned a unique identifier.

        If it has been published before (DLHub detects if it has an identifier), then DLHub
        will update the model to the new version.

        Args:
            model (BaseMetadataModel): Servable to be submitted
        Returns:
            (string) Task ID of this submission, used for checking for success
        """

        # Get the metadata
        metadata = model.to_dict(simplify_paths=True)

        # Mark the method used to submit the model
        metadata['dlhub']['transfer_method'] = {'POST': 'file'}

        # Validate against the servable schema
        validate_against_dlhub_schema(metadata, 'servable')

        # Get the data to be submitted as a ZIP file
        fp, zip_filename = mkstemp('.zip')
        os.close(fp)
        os.unlink(zip_filename)
        try:
            model.get_zip_file(zip_filename)

            # Get the authorization headers
            headers = {}
            self.authorizer.set_authorization_header(headers)

            # Submit data to DLHub service
            with open(zip_filename, 'rb') as zf:
                reply = requests.post(
                    slash_join(self.base_url, 'publish'),
                    headers=headers,
                    files={
                        'json': ('dlhub.json', json.dumps(metadata), 'application/json'),
                        'file': ('servable.zip', zf, 'application/octet-stream')
                    }
                )

            # Return the task id
            if reply.status_code != 200:
                raise Exception(reply.text)
            return reply.json()['task_id']
        finally:
            os.unlink(zip_filename)

    def publish_repository(self, repository):
        """Submit a repository to DLHub for publication

        Args:
            repository (string): Repository to publish
        Returns:
            (string) Task ID of this submission, used for checking for success
        """

        # Publish to DLHub
        metadata = {"repository": repository}
        response = self.post('publish_repo', json_body=metadata)

        task_id = response.data['task_id']
        return task_id



# single source of truth for package version,
# see https://packaging.python.org/en/latest/single_source_version/
__version__ = "0.5.0"

# app name to send as part of SDK requests
app_name = "DLHub SDK v{}".format(__version__)



from dlhub_sdk.utils.types import compose_argument_block, simplify_numpy_dtype
from dlhub_sdk.models.servables import BaseServableModel
import tensorflow as tf
import numpy as np


def _convert_dtype(arg_type):
    """Get a DLHub type for a TensorFlow type

    Args:
        arg_type (int): Enumeration number of a Tensorflow Type
    Returns:
        (string) DLHub-schema-compatible name of the type
    """
    # Get the name of the type
    dtype = tf.DType(arg_type)
    return simplify_numpy_dtype(np.dtype(dtype.as_numpy_dtype))


def _read_tf_inputs_and_outputs(arg_def):
    """Create a DLHub-compatible description from a Google ProtoBuf description of
    the inputs or outputs to a function

    Args:
        arg_def (MessageMap): Description of the inputs/outputs
    Returns:
        - (dict) Description of the input/output data in DLHub format
        - ([string]) Names of the input and output nodes
    """

    # Convert the tensor descriptions to a format compatible with DLHub,
    #   and fetch the names of the corresponding node on the graph
    dlhub_arg_defs = []
    node_names = []
    for name, arg_def in arg_def.items():
        # Get the shape of the Tensor (assuming all inputs are tensors)
        shape = [d.size if d.size != - 1 else None for d in arg_def.tensor_shape.dim]

        # Append the node name
        node_names.append(arg_def.name)

        # Different case if it is a scalar or a tensor
        if len(shape) == 0:
            dlhub_arg_defs.append(compose_argument_block(_convert_dtype(arg_def.dtype), name))
        else:
            dlhub_arg_defs.append(compose_argument_block('ndarray', name,
                                                         shape, _convert_dtype(arg_def.dtype)))

    # If the function has only one argument, return that
    if len(dlhub_arg_defs) == 1:
        return dlhub_arg_defs[0], node_names

    # Otherwise, create a "tuple" type
    #   First sort arguments by description, to ensure a deterministic order to them between runs
    dlhub_arg_defs, node_names = zip(*sorted(zip(dlhub_arg_defs, node_names),
                                             key=lambda x: x[0]['description']))
    return compose_argument_block('tuple', 'Arguments', shape=[len(dlhub_arg_defs)],
                                  element_types=dlhub_arg_defs), list(node_names)


class TensorFlowModel(BaseServableModel):
    """Class for generating descriptions of a TensorFlow model

    Assumes that a model has been saved using the :code:`tf.saved_model` module. Users
    must provide the output directory for the Tensorflow model, and this tool will infer
    all functions that were defined for this model.

    Note: DLHub assumes that the default method for a servable is "run" and TensorFlow assumes
    it to be :code:`tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY`
    (currently defined as "serving_default"). We will rename the default method to "run"
    """

    def _get_type(self):
        return 'TensorFlow Model'

    def _get_handler(self):
        return "tensorflow.TensorFlowServable"

    @classmethod
    def create_model(cls, export_directory):
        """Initialize the desription of a TensorFlow model

        Args:
            export_directory (string): Path to the output directory of a Tensorflow model
        """

        output = cls()

        # Load in the model
        with tf.Session() as sess:
            # Load in the model definition
            model_def = tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING],
                                                   export_directory)

        # Build descriptions for each function in the description
        for name, func_def in model_def.signature_def.items():
            # Get the descriptions for the inputs and outputs
            input_def, input_nodes = _read_tf_inputs_and_outputs(func_def.inputs)
            output_def, output_nodes = _read_tf_inputs_and_outputs(func_def.outputs)

            # Rename the default function
            if name == tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY:
                name = "run"

            # Register the function with the DLHub schema
            output.register_function(name, input_def, output_def,
                                     method_details={'input_nodes': input_nodes,
                                                     'output_nodes': output_nodes})

        # Check if there is a run method
        if 'run' not in output['servable']['methods']:
            raise ValueError('There is no default servable for this model.\n'
                             ' Make sure to use '
                             'tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY '
                             'when saving model.')

        # Add tensorflow version and files
        output.add_requirement('tensorflow', tf.__version__)
        output.add_directory(export_directory)

        return output



from dlhub_sdk.models.servables.python import BasePythonServableModel
from sklearn.base import is_classifier
from sklearn.pipeline import Pipeline
from sklearn.externals import joblib
import sklearn.base as sklbase
import pickle as pkl
import inspect


# scikit-learn stores the version used to create a model in the pickle file,
#  but deletes it before unpickling the object. This code intercepts the version
#  number before it gets deleted

_sklearn_version_global = None
_original_set_state = sklbase.BaseEstimator.__setstate__


def _hijack_baseestimator_setstate(self, state):
    global _sklearn_version_global
    _sklearn_version_global = state.get("_sklearn_version", "pre-0.18")
    _original_set_state(self, state)


# Define the override
sklbase.BaseEstimator.__setstate__ = _hijack_baseestimator_setstate


class ScikitLearnModel(BasePythonServableModel):
    """Metadata for a scikit-learn machine learning model

    This class is build assuming that the inputs to the model will be a list
    of lists of fixed lengths. Models that take different kinds of inputs
    (e.g., Pipelines that include text-processing steps, KernelRidge models
    with custom kernel functions) will are not yet supported.
    """

    @classmethod
    def create_model(cls, path, n_input_columns, classes=None, serialization_method="pickle"):
        """Initialize a scikit-learn model

        Args:
            path (string): Path to model file
            n_input_columns (int): Number of input columns for the model
            classes (Union[int,tuple]): For classification models, number of output classes or a
                list-like object with the names of the classes
            serialization_method (string): Library used to serialize model
        """
        # Load the model and get the method name, needed for instantiating the model type
        skl_version, model = ScikitLearnModel._load_model(path, serialization_method)
        method_name, method_kwargs = ScikitLearnModel._get_predict_method(model)
        output = super(ScikitLearnModel, cls).create_model(method_name, method_kwargs)

        # Save the serialization method
        output["servable"]["options"] = {"serialization_method": serialization_method}

        # Save the classes, if present
        if classes is not None:
            if isinstance(classes, int):
                classes = ['Class {}'.format(i+1) for i in range(classes)]
            output["servable"]["options"]["classes"] = list(classes)

        # Store the model input/output information
        output.n_input_columns = n_input_columns
        output.classes = classes

        # Load other metadata
        output.inspect_model(model)

        # Define the path to the pickle
        output.add_file(path, 'model')

        # Add the sklearn requirement
        output.add_requirement("scikit-learn", skl_version)

        # Set default values for the run operation
        output.set_inputs("ndarray", 'List of records to evaluate with model. Each record is a list'
                                     ' of {} variables.'.format(n_input_columns),
                          item_type='float', shape=[None, n_input_columns])
        if "proba" in method_name:
            output.set_outputs("ndarray", 'Probabilities for membership '
                                          'in each of {} classes'.format(len(classes)),
                               shape=[None, len(classes)], item_type="float")
        else:
            output.set_outputs("ndarray", 'Predictions of the machine learning model.',
                               shape=[None],
                               item_type="float")

        return output

    def _get_type(self):
        return 'Scikit-learn estimator'

    @staticmethod
    def _load_model(path, serialization_method):
        """Load a scikit-learn model from disk to get basic metadata about

        Returns:
            - (string): Scikit-learn version
            - (BaseEstimator) A scikit-learn model object
        """

        # Load in the model
        global _sklearn_version_global
        _sklearn_version_global = None  # Set a default value
        if serialization_method == "pickle":
            with open(path, 'rb') as fp:
                model = pkl.load(fp)
        elif serialization_method == "joblib":
            model = joblib.load(path)
        else:
            raise Exception('Unknown serialization method: {}'.format(serialization_method))

        return _sklearn_version_global, model

    @staticmethod
    def _get_predict_method(model):
        """Get the name of the predict method for this model.

        The method name varies if the model is a classifier or not, and some scikit-learn models
        have arguments for the prediction function

        Args:
            model (BaseEstimator): Model to be inspected
        Returns:
            - (string) Name of the predict method
            - (dict) Any options for the predict method and their default values
        """
        # Store any special keyword arguments for the predict function
        model_obj = model.steps[-1][-1] if isinstance(model, Pipeline) else model
        predict_fun = model_obj.predict_proba if is_classifier(model) else model_obj.predict
        spec = inspect.signature(predict_fun)
        predict_options = dict((k, v.default) for k, v in spec.parameters.items() if k != "X")
        return predict_fun.__name__, predict_options

    def inspect_model(self, model):
        """Extract and store metadata that describes an ML model

        Args:
            model (BaseEstimator): Model to be inspected
        """
        pipeline = isinstance(model, Pipeline)

        # Save the model type
        self._output["servable"]["model_type"] = type(model.steps[-1][-1]).__name__ \
            if pipeline else type(model).__name__

        # Get a summary about the model
        # sklearn prints out a text summary of the model
        self._output["servable"]["model_summary"] = str(model)
        # Determine whether the object is a classifier
        is_clfr = is_classifier(model)
        self["servable"]["options"]["is_classifier"] = is_clfr
        if "classes" not in self["servable"]["options"] and is_clfr:
            raise Exception('Classes (or at least number of classes) must be specified '
                            'in initializer for classifiers.')

    def _get_handler(self):
        return 'sklearn.ScikitLearnServable'



import pkg_resources
import importlib
import requests
import os


from dlhub_sdk.models import BaseMetadataModel


class BaseServableModel(BaseMetadataModel):
    """Base class for servables"""

    def __init__(self):
        super(BaseServableModel, self).__init__()

        # Add the resource type
        #  I chose "InteractiveResource" as the point of DLHub is to provide
        #   web servies for interacting with these models ("query/response portals" are
        #   defined as "InteractiveResources") rather than downloading the source code
        #   (which would fit the definition of software)
        self._output['datacite']['resourceType'] = {'resourceTypeGeneral': 'InteractiveResource'}

        # Initialize the model running-information
        self._output['servable'] = {
            'methods': {'run': {}},
            'shim': self._get_handler(),
            'type': self._get_type(),
            'dependencies': {'python': {}}
        }

    def _get_handler(self):
        """Generate the name of the servable class that DLHub will use to read this metadata

        Returns:
            (string) path to the python class (e.g., "python.PythonStaticMethod")
        """
        raise NotImplementedError()

    def _get_type(self):
        """Get a human-friendly name for this type of servable

        Returns:
            (string): Human-friendly name of an object
        """
        raise NotImplementedError()

    def register_function(self, name, inputs, outputs, parameters=None, method_details=None):
        """Registers a new function to this servable

        See :code:`compose_argument_type` utility function for how to define the inputs
        and outputs to this function.

        Args:
            name (string): Name of the function (e.g., "run")
            inputs (dict): Description of inputs to the function
            outputs (dict): Description of the outputs of the function
            parameters (dict): Any additional parameters for the function and their default values
            method_details (dict): Any options used when constructing a shim to run this function.
        """

        # Check defaults
        if method_details is None:
            method_details = {}
        if parameters is None:
            parameters = {}

        # Add the function
        self._output["servable"]["methods"][name] = {
            'input': inputs,
            'output': outputs,
            'parameters': parameters,
            'method_details': method_details
        }

        return self

    def to_dict(self, simplify_paths=False, save_class_data=False):
        # Make sure the inputs and outputs have been set
        if len(self._output["servable"]["methods"]["run"].get("input", {})) == 0:
            raise ValueError('Inputs have not been defined')
        if len(self._output["servable"]["methods"]["run"].get("output", {})) == 0:
            raise ValueError('Outputs have not been defined')

        return super(BaseServableModel, self).to_dict(simplify_paths)

    def add_requirement(self, library, version=None):
        """Add a required Python library.

        The name of the library should be either the name on PyPI, or a URL for the git repository
        holding the code (e.g., ``git+https://github.com/DLHub-Argonne/dlhub_sdk.git``)

        Args:
            library (string): Name of library
            version (string): Required version. 'latest' to use the most recent version on PyPi (if
                available). 'detect' will attempt to find the version of the library installed on
                the computer running this software. Default is ``None``
        """

        # Attempt to determine the version automatically
        if version == "detect":
            try:
                module = importlib.import_module(library)
                version = module.__version__
            except Exception:
                version = pkg_resources.get_distribution(library).version
        elif version == "latest":
            pypi_req = requests.get('https://pypi.org/pypi/{}/json'.format(library))
            version = pypi_req.json()['info']['version']

        # Set the requirements
        self._output["servable"]["dependencies"]["python"][library] = version
        return self

    def add_requirements(self, requirements):
        """Add a dictionary of requirements

        Utility wrapper for `add_requirement`

        Args:
            requirements (dict): Keys are names of library (str), values are the version
        """
        for p, v in requirements.items():
            self.add_requirement(p, v)
        return self

    def parse_repo2docker_configuration(self, directory=None):
        """Gathers information about required environment from repo2docker configuration files.

        See https://repo2docker.readthedocs.io/en/latest/config_files.html for more details

        Args:
            directory (str): Path to directory containing configuration files
                (default: current working directory)
        """

        # Get a list of all files
        config_files = ['environment.yml', 'requirements.txt', 'setup.py', 'REQUIRE', 'install.R',
                        'apt.txt', 'DESCRIPTION', 'manifest.xml', 'postBuild', 'start',
                        'runtime.txt', 'default.nix', 'Dockerfile']

        # Get the directory name if `None`
        if directory is None:
            directory = os.getcwd()

        # Add every file we can find
        for file in config_files:
            path = os.path.join(directory, file)
            if os.path.isfile(path):
                self.add_file(path)

        return self



"""Model for a pipeline of several servables"""
from dlhub_sdk.models import BaseMetadataModel


class PipelineModel(BaseMetadataModel):
    """Model for a pipeline of several servables

    A Pipeline is created after individual servables have been published in DLHub, or at least
    assigned a DLHub identifier. A Pipeline is formed from a list of these servables, and any
    options to employ when running them. A step in a pipeline could also be another pipeline.

    A simple example for a DLHub pipeline is an image classification tool. The first step in the
    pipeline is an image reader take takes any image type and produces an array. The second step
    standardizes the shape of the image, where the "options" to the servable are the desired
    resolution and whether the image is grayscale or not. The final step is the classification
    pipeline. Put together, the image pipeline can support any type of input data.
    """

    def __init__(self):
        super(PipelineModel, self).__init__()

        # Make as a "InteractiveResource"
        self['datacite']['resourceType'] = {'resourceTypeGeneral': 'InteractiveResource'}

        # Add list of pipeline steps
        self._output['pipeline'] = {'steps': []}

    def add_step(self, author, name, description, parameters=None):
        """Add a step to the pipeline

        Args:
            author (string): DLHub username of the owner of the servable
            name (string): Name of the DLHub servable
            description (string): A short description of this step
            parameters (dict): Any options for the servable.
                See the list of parameters for a servable
        """

        # Compose the block
        step = {
            'author': author,
            'name': name,
            'description': description
        }
        if parameters is not None:
            step['parameters'] = parameters
        self["pipeline"]["steps"].append(step)



import os
from setuptools import setup, find_packages

# single source of truth for package version
version_ns = {}
with open(os.path.join("dlhub_sdk", "version.py")) as f:
    exec(f.read(), version_ns)
version = version_ns['__version__']

setup(
    name='dlhub_sdk',
    version=version,
    packages=find_packages(),
    description='Python interface and utilities for DLHub',
    long_description=("DLHub SDK contains a Python interface to the Data "
                      "and Learning Hub for Science (DLHub). These interfaces "
                      "include functions for quickly describing a model in the "
                      "correct schema for DLHub, and discovering or using models "
                      "that other scientists have published."),
    install_requires=[
        "pandas", "requests", "jsonschema", "globus_sdk", "jsonpickle", "mdf_toolbox"
    ],
    python_requires=">=3.4",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    keywords=[
        "DLHub",
        "Data and Learning Hub for Science",
        "machine learning",
        "data publication",
        "reproducibility",
    ],
    author='Ben Blaiszik',
    author_email='bblaiszik@anl.gov',
    license="Apache License, Version 2.0",
    url="https://github.com/DLHub-Argonne/dlhub_sdk"
)



from keras import __version__ as keras_version
from keras.models import load_model

from dlhub_sdk.models.servables.python import BasePythonServableModel
from dlhub_sdk.utils.types import compose_argument_block


_keras_version_tuple = tuple(int(i) for i in keras_version.split("."))


class KerasModel(BasePythonServableModel):
    """Servable based on a Keras Model object.

    Assumes that the model has been saved to an hdf5 file"""

    @classmethod
    def create_model(cls, model_path, output_names):
        """Initialize a Keras model.

        Args:
            model_path (string): Path to the hd5 file that describes a model and the weights
            output_names ([string] or [[string]]): Names of output classes.
                If applicable, one list for each output layer.
       """
        output = super(KerasModel, cls).create_model('predict')

        # Add model as a file to be sent
        output.add_file(model_path, 'model')

        # Get the model details
        model = load_model(model_path)

        # Get the inputs of the model
        output['servable']['methods']['run']['input'] = output.format_layer_spec(model.input_shape)
        output['servable']['methods']['run']['output'] = output.format_layer_spec(
                                                                    model.output_shape)
        output['servable']['methods']['run']['method_details']['classes'] = output_names

        # Get a full description of the model
        output.summary = ""

        def capture_summary(x):
            output.summary += x + "\n"
        model.summary(print_fn=capture_summary)
        output['servable']['model_summary'] = output.summary
        output['servable']['model_type'] = 'Deep NN'

        # Add keras as a dependency
        output.add_requirement('keras', keras_version)
        output.add_requirement('h5py', 'detect')
        return output

    def format_layer_spec(self, layers):
        """Make a description of a list of input or output layers

        Args:
            layers (tuple or [tuple]): Shape of the layers
        Return:
            (dict) Description of the inputs / outputs
        """
        if isinstance(layers, tuple):
            return compose_argument_block("ndarray", "Tensor", shape=list(layers))
        else:
            return compose_argument_block("tuple", "Tuple of tensors",
                                          element_types=[self.format_layer_spec(i) for i in layers])

    def _get_handler(self):
        return "keras.KerasServable"

    def _get_type(self):
        return "Keras Model"



"""Tools to annotate generic operations (e.g., class method calls) in Python"""
import pickle as pkl

from dlhub_sdk.models.servables import BaseServableModel
from dlhub_sdk.utils.types import compose_argument_block


class BasePythonServableModel(BaseServableModel):
    """Describes a static python function to be run"""

    def __init__(self):
        super(BasePythonServableModel, self).__init__()

        # Update some information about the servable
        self._output['servable'].update({
            'language': 'python',
        })

    @classmethod
    def create_model(cls, method, function_kwargs=None):
        """Initialize a model for a python object

        Args:
            method (string): Name of the method for this class
            function_kwargs (dict): Names and values of any other argument of the function to set
                the values must be JSON serializable.
        """
        output = cls()

        # Get default values
        if function_kwargs is None:
            function_kwargs = dict()

        # Set values
        output.register_function("run", {}, {}, function_kwargs, {'method_name': method})
        return output

    def set_inputs(self, data_type, description, shape=(), item_type=None, **kwargs):
        """Define the inputs to the default ("run") function

        Args:
            data_type (string): Type of the input data
            description (string): Human-friendly description of the data
            shape (list): Required for data_type of list or ndarray. Use `None` for dimensions that
                can have any numbers of values
            item_type (string/dict): Description of the item type. Required for data_type = list
        """
        args = compose_argument_block(data_type, description, shape, item_type, **kwargs)

        # Set the inputs
        self._output["servable"]["methods"]["run"]["input"] = args
        return self

    def set_outputs(self, data_type, description, shape=(), item_type=None, **kwargs):
        """Define the outputs to the default ("run") function

        Args:
            data_type (string): Type of the output data
            description (string): Human-friendly description of the data
            shape (list): Required for data_type of ndarray. Use `None` for dimensions that
                can have any numbers of values
            item_type (string): Description of the type of item in a list
        """

        args = compose_argument_block(data_type, description, shape, item_type, **kwargs)
        self._output["servable"]["methods"]["run"]["output"] = args
        return self


class PythonClassMethodModel(BasePythonServableModel):
    """Model for describing servables where the function to be performed is a method of a class.

    To use this model, you must define the path to the pickled object and the function of that
    object to be called. Any additional libraries (beyond the standard libraries) required
    to run the library and their versions must also be specified. You may also specify
    any arguments of the class that should be set as defaults."""

    def __init__(self):
        super(PythonClassMethodModel, self).__init__()
        self.class_name = None

    @classmethod
    def create_model(cls, path, method, function_kwargs=None):
        """Initialize a model for a python object

        Args:
            path (string): Path to a pickled Python file
            method (string): Name of the method for this class
            function_kwargs (dict): Names and values of any other argument of the function to set
                the values must be JSON serializable.
        """
        output = super(PythonClassMethodModel, cls).create_model(method, function_kwargs)

        output.add_file(path, 'pickle')

        # Get the class name
        with open(path, 'rb') as fp:
            obj = pkl.load(fp)
            class_name = '{}.{}'.format(obj.__class__.__module__, obj.__class__.__name__)

        output._output["servable"]["methods"]["run"]["method_details"].update({
            'class_name': class_name
        })

        return output

    def _get_handler(self):
        return 'python.PythonClassMethodServable'

    def _get_type(self):
        return 'Python class method'


class PythonStaticMethodModel(BasePythonServableModel):
    """Model for a servable that calls a Python static method. Static methods can be called
    without any other context, unlike the class methods in PickledClassServableModelBase.

    An example static method is the sqrt operation from numpy, `numpy.sqrt`. You can make a model
    of this function by calling :code:`PythonStaticMethodModel.from_function_pointer(numpy.sqrt)`.
    """

    def __init__(self):
        super(PythonStaticMethodModel, self).__init__()
        self.module = None
        self.autobatch = False

    @classmethod
    def create_model(cls, module, method, autobatch=False, function_kwargs=None):
        """Initialize the method

        Args:
            module (string): Name of the module holding the function
            method (string): Name of the method for this class
            autobatch (bool): Whether to automatically run this function on a list of inputs.
                Calls :code:`map(f, list)`
            function_kwargs (dict): Names and values of any other argument of the function to set
                the values must be JSON serializable.
        """
        output = super(PythonStaticMethodModel, cls).create_model(method, function_kwargs)

        output._output["servable"]["methods"]["run"]["method_details"].update({
            'module': module,
            'autobatch': autobatch
        })
        return output

    @classmethod
    def from_function_pointer(cls, f, autobatch=False, function_kwargs=None):
        """Construct the module given a function pointer

        Args:
            f (object): A function pointer
            autobatch (bool): Whether to run function on an iterable of entries
            function_kwargs (dict): Any default options for this function
        """
        return cls.create_model(f.__module__, f.__name__, autobatch, function_kwargs)

    def _get_handler(self):
        return 'python.PythonStaticMethodServable'

    def _get_type(self):
        return 'Python static method'



"""Utilities for generating descriptions of data types"""
from six import string_types


def simplify_numpy_dtype(dtype):
    """Given a numpy dtype, write out the type as string

    Args:
        dtype (numpy.dytpe): Type
    Returns:
        (string) name as a simple string
    """

    kind = dtype.kind
    if kind == "b":
        return "boolean"
    elif kind == "i" or kind == "u":
        return "integer"
    elif kind == "f":
        return "float"
    elif kind == "c":
        return "complex"
    elif kind == "m":
        return "timedelta"
    elif kind == "M":
        return "datetime"
    elif kind == "S" or kind == "U":
        return "string"
    else:
        return "python object"


def compose_argument_block(data_type, description, shape=(), item_type=None,
                           python_type=None, properties=None, element_types=None, **kwargs):
    """Compile a list of argument descriptions into an argument_type block

    Args:
        data_type (string): Type of the input data
        description (string): Human-friendly description of the data
        shape (list): Required for data_type of list or ndarray. Use `None` for dimensions that
            can have any numbers of values
        item_type (string/dict): Description of the item type. Required for data_type = list.
            Can either be a string type, or a dict that is a valid type for an argument type block
        python_type (string): Full path of a Python object type
            (e.g., :code:`pymatgen.core.Compostion`)
        properties (dict): Descriptions of the types in a dictionary
        element_types ([dict] or [list]): Types of elements in a tuple. List of type definition
            dictionaries or types as strings.
    Keyword Arguments: Any other details particular to this kind of data
    Returns:
        (dict) Description of method in a form compatible with DLHub
    """
    # Initialize the description
    args = {
        'type': data_type,
        'description': description
    }

    # Check that shape is specified if need be
    if data_type == "ndarray":
        if len(shape) == 0:
            raise ValueError('Shape must be specified for ndarrays')
        args['shape'] = list(shape)

    # Check if the item_type needs to be defined
    if data_type == "list":
        if item_type is None:
            raise ValueError('Item type must be defined for lists')

    # Check if the element_type need be defined
    if data_type == "tuple":
        if element_types is None:
            raise ValueError('Element type must be defined for tuples')

    # If the type is "python object", python_type must be specified
    if data_type == "python object":
        if python_type is None:
            raise ValueError('Python type must be defined ')
        args['python_type'] = python_type

    # Check if the keys need to be defined
    if data_type == "dict":
        if properties is None:
            raise ValueError('Properties must be defined for dict type')
        args['properties'] = properties

    # Define the item types
    if item_type is not None:
        if isinstance(item_type, dict):
            args['item_type'] = item_type
        elif isinstance(item_type, string_types):  # Is a string
            args['item_type'] = {'type': item_type}

    # Define the types of tuples
    if element_types is not None:
        args['element_types'] = list(element_types)

    # Add in any kwargs
    args.update(**kwargs)
    return args
